class TAGS:
    """
    This class contains constants for different tag types used in the system.
    """

    AUTH = "auth"
    PROJECT = "project"
    FEATURE_STORE = "feature_store"
    FEATURE_GROUP = "feature_group"
    FEATURE_VIEW = "feature_view"
    DATASET = "dataset"
    ENVIRONMENT = "environment"
    JOB = "job"
    SYSTEM = "system"

    STATEFUL = "stateful"
    STATELESS = "stateless"

    READ = "read"
    WRITE = "write"
